/*
	e2floppy.cc

	flexemu, an MC6809 emulator running FLEX
	Copyright (C) 1997  W. Schwotzer

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "misc1.h"

#include <stdlib.h>
#include <string.h>
#ifdef _MSC_VER
#include <io.h>
#endif
#include "e2floppy.h"

// supported DSK formats used by Ben Williams FLEX Emulator:
// (must be sorted in increasing disk size order!)
struct s_formats formats[] = {
	    { 89600L,  35, 10, 6  },
	    { 102400L, 40, 10, 6  },
	    { 179200L, 35, 20, 16 },
	    { 204800L, 40, 20, 16 },
	    { 358400L, 35, 40, 36 },
	    { 409600L, 40, 40, 36 },
	    { 819200L, 80, 40, 36 },
	    { 0L,      0, 0, 0 }
};

E2floppy::E2floppy(Inout *x_io, Mc6809 *x_cpu)
{
	Word i;

	cpu = x_cpu;
	io  = x_io;

	disk_dir = "";
	for (i = 0; i < 4; i++)
		header_for_no_format(i);
	selected = 4;			// deselect all drives
	pfs = &floppy[selected];	// set pfs to selected drive
	pfs->ready = 0; 		// drive number 4 is always not ready
} // E2floppy

void E2floppy::header_for_no_format(Word drive_nr)
{
	s_floppy *pfloppy;

	if (drive_nr > 3)
		return;
	pfloppy = &floppy[drive_nr];	// set pfs to selected drive
	pfloppy->type  = NO_FORMAT;
	pfloppy->fp    = NULL;
	pfloppy->ppath = NULL;
	pfloppy->track = 1;	// position to a track != 0  !!!
	pfloppy->ready = 0;
}

void E2floppy::header_for_own_format(
	s_floppy	*pfloppy,
	s_flex_header	*pheader,
	Byte		wp)
{
	floppy->ready          = 1;
	pfloppy->type          = FLX_FORMAT;
	pfloppy->offset        = sizeof(struct s_flex_header);
	pfloppy->write_protect =
		(wp | pheader->write_protect) ? 0x40 : 0;
	pfloppy->max_sector    = pheader->sectors;
	pfloppy->max_sector0   = pheader->sectors0;
	pfloppy->max_track     = pheader->tracks - 1;
	pfloppy->byte_p_sector = 128 << pheader->sizecode;
	pfloppy->byte_p_track0 =
		pheader->sides0 * pheader->sectors0 * floppy->byte_p_sector;
	pfloppy->byte_p_track  =
		pheader->sides  * pheader->sectors  * floppy->byte_p_sector;
} // header_for_own_format

void E2floppy::header_for_ben_format(
	s_floppy	*pfloppy,
	s_formats	*pformat,
	Byte		wp)
{
	pfloppy->ready         = 1;
	pfloppy->type          = DSK_FORMAT;
	pfloppy->offset        = 0;
	pfloppy->write_protect = wp;
	pfloppy->max_sector    = pformat->sectors >> 1;
	pfloppy->max_sector0   = pformat->sectors >> 1;
	pfloppy->max_track     = pformat->tracks - 1;
	pfloppy->byte_p_sector = SECTOR_SIZE;
	pfloppy->byte_p_track0 = pformat->sectors << 8; 
	pfloppy->byte_p_track  = pformat->sectors << 8; 
} // header_for_ben_format


// return != 0 on success
Byte E2floppy::umount_drive(Word drive_nr)
{
	struct s_floppy *pfloppy;

	if (drive_nr > 3)
		return 0;
	pfloppy = &floppy[drive_nr];
	if (pfloppy->ready == 0)
		return 0;
#ifdef NAFS
	if (pfloppy->type == NAFS_FORMAT) {
		umount_nafs(drive_nr);
	}
#endif
	if (pfloppy->type == FLX_FORMAT || pfloppy->type == DSK_FORMAT) {
		if (pfloppy->fp != NULL)
			fclose(pfloppy->fp);
	}
	if (pfloppy->ppath != NULL) {
		free(pfloppy->ppath);
	}
	header_for_no_format(drive_nr);
	return 1;
} // umount_drive

// return != 0 on success
Byte E2floppy::mount_drive(char *path, Word drive_nr)
{
	struct s_floppy	*pfloppy ;
	char		*p;
	Byte 		 result;

	if (drive_nr > 3 || path == NULL || strlen(path) == 0)
		return 0;
	pfloppy = &floppy[drive_nr];
	if (pfloppy->ready != 0)
		return 0;
	if ((p = (char *)malloc(strlen(path) + 1)) == NULL)
		return 0;
	strcpy(p, path);
	pfloppy->track = 1;	// position to a track != 0  !!!
	pfloppy->ppath = p;
	if((result = init_disk_image(pfloppy)) == 0)
#ifdef NAFS
		if((result = init_nafs_drive(pfloppy)) == 0)
#endif // NAFS
	{
		free(p);
		pfloppy->ppath = NULL;
		if ((p = (char *)malloc(strlen(disk_dir) +
			strlen(path) + 2)) == NULL)
			return 0;
		strcpy(p, disk_dir);
		if (strlen(disk_dir) > 0 &&
			*(disk_dir + strlen(disk_dir) - 1) != PATHSEPARATOR)
			strcat(p, PATHSEPARATORSTRING);
		strcat(p, path);
		pfloppy->ppath = p;
		if((result = init_disk_image(pfloppy)) == 0)
#ifdef NAFS
			if((result = init_nafs_drive(pfloppy)) == 0)
#endif // NAFS
			{
				free(p);
				pfloppy->ppath = NULL;
			}
	}
	return result;
} // mount_drive

void E2floppy::create_boot_sector(Byte sec_buf[])
{
	FILE	*boot;
	int	i;

	// first write boot sector if present as file "boot"
	sec_buf[0] = 0x39;	// means RTS
	for (i = 1; i < SECTOR_SIZE; i++)
		sec_buf[i] = 0;
#ifdef sun
	if((boot = fopen("boot", "r")) != NULL)
#else
	if((boot = fopen("boot", "rb")) != NULL)
#endif
		fread(sec_buf, SECTOR_SIZE, 1, boot);
} // create_boot_sector

/* ARGSUSED */
void E2floppy::create_sector2(Byte sec_buf[], struct s_formats *fmt)
{
	int	i;

	// create unused (???) sector 2
	for (i = 0; i < SECTOR_SIZE; i++)
		sec_buf[i] = 0;
	sec_buf[1] = 3;	// link to next sector
} // create_sector2

void E2floppy::create_sys_info_sector(Byte sec_buf[], char *name,
	struct s_formats *fmt)
{
	struct s_sys_info_sector *sys;
	int		i, start, free;
	time_t		time_now;
	struct tm	*lt;

	for (i = 0; i < SECTOR_SIZE; i++)
		sec_buf[i] = 0;
	sys = (struct s_sys_info_sector *)sec_buf;
	for (i = 0; i < 8; i++) {
		if (*(name+i) == '.' || *(name+i) == '\0')
			break;
		sys->disk_name[i] = *(name+i);
	} // for
	start			= fmt->dir_sectors + 4;
	free			= (fmt->sectors * fmt->tracks) - start;
	time_now		= time(NULL);
	lt			= localtime(&time_now);
	sys->disk_number[0]	= 0;
	sys->disk_number[1]	= 1;
	sys->fc_start_trk	= start / fmt->sectors;
	sys->fc_start_sec	= (start % fmt->sectors) + 1;
	sys->fc_end_trk		= fmt->tracks - 1;
	sys->fc_end_sec		= (Byte)fmt->sectors;
	sys->free[0]		= free >> 8;
	sys->free[1]		= free & 0xff;
	sys->month		= lt->tm_mon + 1;
	sys->day		= lt->tm_mday;
	sys->year		= lt->tm_year;
	sys->last_trk		= fmt->tracks - 1;
	sys->last_sec		= (Byte)fmt->sectors;
} // create_sys_info_sectors

void E2floppy::write_dir_sectors(FILE *fp, struct s_formats *fmt)
{
	Byte		sec_buf[SECTOR_SIZE];
	int		i;

	for (i = 0; i < SECTOR_SIZE; i++)
		sec_buf[i] = 0;
	for (i = 0; i < fmt->dir_sectors; i++) {
		sec_buf[0] = 0;
		sec_buf[1] = 0;
		if (i < fmt->dir_sectors - 1) {
			sec_buf[0] = (i + 5) / fmt->sectors;
			sec_buf[1] = ((i + 5) % fmt->sectors) + 1;
		}
		fwrite(sec_buf, SECTOR_SIZE, 1, fp);
	}
} // write_dir_sectors

void E2floppy::write_sectors(FILE *fp, struct s_formats *fmt)
{
	Byte		sec_buf[SECTOR_SIZE];
	int		i;

	for (i = 0; i < SECTOR_SIZE; i++)
		sec_buf[i] = 0;
	for (i = fmt->dir_sectors + 5;
		i <= fmt->sectors * fmt->tracks; i++) {
		sec_buf[0] = 0;
		sec_buf[1] = 0;
		if (i != fmt->sectors * fmt->tracks) {
			sec_buf[0] = i / fmt->sectors;
			sec_buf[1] = (i % fmt->sectors) + 1;
		}
		fwrite(sec_buf, SECTOR_SIZE, 1, fp);
	}
} // write_sectors

// return != 0 on success
// size given in KByte
Byte E2floppy::format_disk(Word type, SDWord size, char *name)
{
	FILE		*fp;
	char		*path;
	struct s_formats *fmt;
	Byte		sector_buffer[SECTOR_SIZE];

	if (disk_dir == NULL || strlen(disk_dir) == 0 ||
		name == NULL || strlen(name) == 0)
		return 0;
	fmt = formats;
	while ((size * 1024) > fmt->size) {
		if ((++fmt)->size == 0) {
			fmt--; // at least use the largest format
			break;
		}
	}
	path = (char *)malloc(strlen(disk_dir) + strlen(name) + 1);
	if (path == NULL)
		return 0;
	strcpy(path, disk_dir);
	if (strlen(path) > 0 && path[strlen(path)-1] != PATHSEPARATOR)
		strcat(path, PATHSEPARATORSTRING);
	strcat(path, name);
#ifdef sun
	fp = fopen(path, "w");
#else
	fp = fopen(path, "wb");
#endif
	free(path);
	if (fp != NULL) {
		create_boot_sector(sector_buffer);
		fwrite(sector_buffer, SECTOR_SIZE, 1, fp);	
		create_sector2(sector_buffer, fmt);
		fwrite(sector_buffer, SECTOR_SIZE, 1, fp);	
		create_sys_info_sector(sector_buffer, name, fmt);
		fwrite(sector_buffer, SECTOR_SIZE, 1, fp);	
		fwrite(sector_buffer, SECTOR_SIZE, 1, fp);	
		write_dir_sectors(fp, fmt);
		write_sectors(fp, fmt);
		fclose(fp);
	} // if

	return 1;
} // format_disk

void E2floppy::disk_directory(char *x_disk_dir)
{
	disk_dir = x_disk_dir;
}

void E2floppy::mount_all_drives(char *drive[])
{
	int i;

	for (i = 0; i < 4; i++)
		mount_drive(drive[i], i);
	selected = 4;			// deselect all drives
	pfs = &floppy[selected];	// set pfs to selected drive
	pfs->ready = 0; 		// drive number 4 is always not ready
}  // mount_all_drives

Byte E2floppy::umount_all_drives(void)
{
	    Word i;
	    Byte result;

	result = 1;
	for (i = 0; i < 4; i++)
		if (!umount_drive(i))
			result = 0;
	return result;
}  // umount_all_drives

// get info for corresponding drive or NULL
// the info string should not exceed 512 Bytes
// it is dynamically allocated and should be freed
// by the calling program

char *E2floppy::drive_info(Word drive_nr)
{
	char		*pstr, *ptype;
	struct s_floppy	*pfl;

	pstr = NULL;
	if (drive_nr > 3 ||
		((pstr = (char *)malloc(512)) == NULL))
		return pstr;
	pfl = &floppy[drive_nr];
	if (pfl->ready == 0) {
		sprintf(pstr, "drive #%d not ready\n", drive_nr);
		return pstr;
	} else {
	switch (pfl->type) {
	   case FLX_FORMAT: ptype = "disk image (FLX-Format)"; break;
	   case DSK_FORMAT: ptype = "disk image (DSK-Format)"; break;
#ifdef NAFS
	   case NAFS_FORMAT: ptype = "Directory"; break;
#endif
	   default: ptype = "unspecified"; break;
	}
	sprintf(pstr,
"drive       #%d\n"
"format:     %s\n"
"path:       %s\n"
"tracks:     %d\n"
"sectors:    %d\n"
"write-prot: %s\n", drive_nr, ptype, pfl->ppath,
	pfl->max_track + 1, 2 * pfl->max_sector,
	pfl->write_protect ? "yes" : "no");
	}
	return pstr;
} // drive_info

char *E2floppy::open_mode(char *path)
{
	    int  wp;	// write protect flag
	char *mode;
	
	wp = access(path, W_OK);
#ifdef sun
	mode = wp ? "r" : "r+";
#else
	mode = wp ? "rb" : "rb+";
#endif
	return mode;
} // open_mode


Byte E2floppy::init_disk_image(struct s_floppy *pfloppy)
{
	int			wp;	// write protect flag
	struct stat		sbuf;
	struct s_flex_header	header;
	struct s_formats	*pformat;

	if(stat(pfloppy->ppath, &sbuf))
		return 0;
#ifdef _MSC_VER
	if (sbuf.st_mode & _S_IFREG) {
#else
	if (S_ISREG(sbuf.st_mode)) {
#endif
		wp = access(pfloppy->ppath, W_OK);
		if ((pfloppy->fp = fopen(pfloppy->ppath,
			open_mode(pfloppy->ppath))) == NULL)
			return 0;
		if ((fread((void *)&header, sizeof(struct s_flex_header), 1,
		   pfloppy->fp) == 1) && 
		   (header.magic_number == MAGIC_NUMBER)) {
		   	header_for_own_format(pfloppy, &header, wp);
			return 1;	// it's a disk image with own format 
		} // if
		pformat = formats;
		while (pformat->size) {
			if (pformat->size == sbuf.st_size) {
				header_for_ben_format(pfloppy, pformat, wp);
				return 1; // it's ben's format 
			}
			pformat++; 
		} // while
	}  // if
	return 0;
} // init_disk_image


E2floppy::~E2floppy(void)
{
	int i;

	for (i = 0; i < 4; i++) {
		if (floppy[i].fp != NULL) {
			fclose(floppy[i].fp);
			floppy[i].fp = NULL;
		}
		if (floppy[i].ppath != NULL) {
			free(floppy[i].ppath);
			floppy[i].ppath = NULL;
		}
#ifdef NAFS
		umount_nafs(i);
#endif
	}
} // ~E2floppy

Byte E2floppy::update_all_drives(void)
{
	struct s_floppy	*pfloppy;
	Word		i;
	Byte		result;

	result = 1;
	for (i = 0; i < 4; i++) {
		pfloppy = &floppy[i];
		if (pfloppy->ready == 0)
			// no error if drive not ready
			continue;
		if (!update_drive(i))
			result = 0;
	}
	return result;
} // update_all_drives

Byte E2floppy::update_drive(Word drive_nr)
{
	struct s_floppy *pfloppy;

	if (drive_nr > 3)
		return 0;
	pfloppy = &floppy[drive_nr];
	if (pfloppy->ready == 0)
		// error if drive not ready
		return 0;
#ifdef NAFS
	if (pfloppy->type == NAFS_FORMAT) {
		if (open_files(drive_nr))
			return 0;
		umount_nafs(drive_nr);
		mount_nafs(pfloppy, drive_nr);
		pfloppy->track = 1;
	} // if
#endif
	return 1;
} // update_drive

void E2floppy::resetIo()
{
	drisel = 0;
	Wd1793::resetIo();
}


Byte E2floppy::readIo(Word offset)
{
	if (offset <= 3)
		return Wd1793::readIo(offset);
	status = 0xff;	// unused is logical high
	if (!side)
		status &= 0xfd;
	if (!irq)
		status &= 0xbf;
	if (!drq)
		status &= 0x7f;
	return status;
} // readIo


void E2floppy::writeIo(Word offset, Byte val)
{
	if (offset <= 3)
		Wd1793::writeIo(offset, val);
	else {
		drisel = val;
		side = (drisel & 0x10) ? 1 : 0;
		switch (drisel & 0x0f) {
			case 0x01 : selected = 0;
				    break;
			case 0x02 : selected = 1;
				    break;
			case 0x04 : selected = 2;
				    break;
			case 0x08 : selected = 3;
				    break;
			default   : selected = 4;
		};
		pfs->track = tr;
		pfs = &floppy[selected];
		tr = (Byte)pfs->track;
	}
} // writeIo


Byte E2floppy::positionDrive(void)
{
	long fpos; // must be long for fseek

	if (!tr)
		fpos = pfs->offset +
			((sr - 1) * pfs->byte_p_sector);
	else
		fpos = pfs->offset +
			pfs->byte_p_track0 +
			((sr-1) * pfs->byte_p_sector) +
			((tr-1) * pfs->byte_p_track);
	return fseek(pfs->fp, fpos, SEEK_SET);
}

Byte E2floppy::readByte(Word index)
{

	if (index == pfs->byte_p_sector) {
#ifdef NAFS
		if (pfs->type != NAFS_FORMAT) {
#endif
			if (positionDrive() ||
			   (fread((void *)&sector_buffer, pfs->byte_p_sector, 1, 
			   pfs->fp) != 1)) {
				drq = 0;
				str = 0x10;
				byteCount = 0;
			}
#ifdef NAFS
		} else
			read_nafs_sector();
#endif
	} // if
	return sector_buffer[SECTOR_SIZE - index];
} // readByte


void E2floppy::writeByte(Word index)
{
	unsigned int error;

	sector_buffer[SECTOR_SIZE - index] = dr;
	if (index == 1) {
#ifdef NAFS
		if (pfs->type != NAFS_FORMAT) {
#endif
			error =  positionDrive();
			error |= (fwrite((void *)&sector_buffer, pfs->byte_p_sector, 1, 
				pfs->fp) != 1);
#ifdef NAFS
		} else
			write_nafs_sector();
#endif // NAFS
	} // if
} // writeByte


Byte E2floppy::recordNotFound(void)
{
	if (tr) 
		return (sr == 0 || sr > (pfs->max_sector * 2)); 
	else
		return (sr == 0 || sr > (pfs->max_sector0 * 2)); 
} // recordNotFound

Byte E2floppy::seekError(Byte new_track)
{
	return (new_track > pfs->max_track);
} // recordNotFound

Byte E2floppy::driveReady(void)
{
	return pfs->ready;
}  // driveReady

Byte E2floppy::writeProtect(void)
{
	return pfs->write_protect;
}  // writeProtect

