/*
    main.cc


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "misc1.h"

#ifndef _MSC_VER
#include <unistd.h>
#endif
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <limits.h>
#include <time.h>
#ifdef WIN32
#include <windows.h>
#ifndef PATH_MAX
#define PATH_MAX MAX_PATH
#endif
#endif

#include "classes.h"
#include "e2.h"
#include "mc6809.h"
#include "inout.h"
#include "e2video.h"
#include "pia2.h"
#include "da6809.h"
#include "acia1.h"
#include "mc146818.h"
#include "e2floppy.h"
#include "command.h"
#include "mmu.h"
#include "pia1.h"
#include "win32gui.h"


// class memory is now included in cpu:
#define MEMORY cpu
#define PMEMORY (*MEMORY)

#ifdef UNIX
#ifdef __osf__
extern "C" unsigned int alarm(unsigned int);
#endif
#endif

#ifndef linux
extern int optind;
extern int opterr;
extern char *optarg;
#endif

struct sOptions {
	char	 *drives[4];
	char	*hex_file;
	char	*disk_dir;
	Byte	 term_mode;
	short int reset_key; // must be short int because of sscanf !!!
};

#ifdef WIN32
UINT idTimer;	// id of global system timer
#endif

#ifndef WIN32
#ifdef SIGALRM
extern "C" void update(int)
{
	Mc6809::update_flag = 1;
	(void)signal(SIGALRM, update);
	alarm(1);
}
#endif // SIGALRM
#else
VOID CALLBACK OneSecTimerProc(HWND hwnd, UINT uMsg, UINT idEvent, DWORD dwTime)
{
	Mc6809::update_flag = 1;
} // OneSecTimerProc
#endif // WIN32

#ifdef WIN32
// uses its own implementation of getopt
int optind = 1;
int opterr = 0;
char *optarg = NULL;

int getopt(int argc, char *argv[], char *optstr)
{
	int	i;
	char	opt;

	while (1) {
		optarg = optstr;
		for (i = 1; i < optind; i++) {
 			if (*(++optarg) == ':')
				optarg++;
		}
		if ((opt = *optarg) == '\0')
			return -1;
		optind++;
		i = 1;
		while (i < argc) {
			if (argv[i][0] == '-' && argv[i][1] == opt) {
				// found option
				if (*(optarg + 1) == ':') {
					// option has a parameter
					if (argv[i][2] != '\0') {
						optarg = &argv[i][2];
						return opt;
					} else {
						if (++i < argc) {
							optarg = &argv[i][0];
							return opt;
						} else
							break;
					} // else
				} else
					// option has no parameter
					return opt;
			} // if
			i++;
		} // while
	} // while
	return -1;
} // getopt
#endif


#ifndef WIN32
void print_help(void) {
	fprintf(stderr, "usage: flexemu <options>\n");
	fprintf(stderr, "  <options> are:\n");
	fprintf(stderr, "  -f <hexfile>\n");
	fprintf(stderr, "  -0 <diskimage/directory drive 0>\n");
	fprintf(stderr, "  -1 <diskimage/directory drive 1>\n");
	fprintf(stderr, "  -2 <diskimage/directory drive 2>\n");
	fprintf(stderr, "  -3 <diskimage/directory drive 3>\n");
	fprintf(stderr, "  -p (directory for FLEX disks)\n");
	fprintf(stderr, "  -j <factor for screen width>\n");
	fprintf(stderr, "  -k <factor for screen height>\n");
#ifdef UNIX
	fprintf(stderr, "  -t (terminal only mode)\n");
	fprintf(stderr, "  -r <two-hex-digit reset key>\n");
#endif
#ifdef X11
	fprintf(stderr, "  -s (run X11 synchronized)\n");
	fprintf(stderr, "  -c <color> define foreground color\n");
	fprintf(stderr, "  -i (display inverse video)\n");
#endif
	fprintf(stderr, "  -h (display this)\n");
	fprintf(stderr, "  -? (display this)\n");
	fprintf(stderr, "  -v (print version number)\n");
} // print_help
#endif // WIN32

void init_options(struct sGuiOptions *pGuiOptions,
			struct sOptions *pOptions,
			int argc,
			char **argv)
{
	pOptions->drives[0] = "system.dsk";
	pOptions->drives[1] = "";
	pOptions->drives[2] = "";
	pOptions->drives[3] = "";
	pOptions->disk_dir  = "";
	pOptions->hex_file  = "neumon54.hex";
	pOptions->term_mode        = 0;
	pOptions->reset_key        = 0x1e; // is Ctrl-^ for reset or Sig. INT

	pGuiOptions->argc          = argc;
	pGuiOptions->argv          = argv;
	pGuiOptions->color         = "green";
	pGuiOptions->inverse       = 0;
	pGuiOptions->www_browser   = "";
	pGuiOptions->doc_dir       = "";
	pGuiOptions->guiXSize      = 1;
	pGuiOptions->guiYSize      = 1;
	pGuiOptions->synchronized  = 0;
	// automatic SERPAR switch only if Eurocom II Monitorprogram loaded
	pGuiOptions->switch_sp = !strcmp(pOptions->hex_file, "neumon54.hex");
} // init_options

void get_options(struct sGuiOptions *pGuiOptions,
			struct sOptions *pOptions,
			int argc,
			char **argv)
{
	int	result;
	char	optstr[32];
	int	i;
#ifndef WIN32
	char	envName[10], *p;

	// first look for environment variables
	if ((p = getenv(FLEXINVERSE)) != NULL && stricmp(p, "true") == 0)
		pGuiOptions->inverse = 1;
	if ((p = getenv(FLEXWWWBROWSER)) != NULL)
		pGuiOptions->www_browser = p;
	if ((p = getenv(FLEXDOCDIR)) != NULL)
		pGuiOptions->doc_dir = p;
	if ((p = getenv(FLEXMONITOR)) != NULL)
		pOptions->hex_file = p;
	if ((p = getenv(FLEXCOLOR)) != NULL)
		pGuiOptions->color = p;
	if ((p = getenv(FLEXDISKDIR)) != NULL)
		pOptions->disk_dir = p;
	strcpy(envName, FLEXDISKX);
	for (i = 0; i <= 3; i++) {
		envName[8] = i + '0';
		if ((p = getenv(envName)) != NULL)
			pOptions->drives[i] = p;
	}
#endif  // ifndef WIN32
	optind = 1;
	opterr = 0;
	strcpy((char *)optstr, "p:f:0:1:2:3:j:k:");
#ifdef UNIX
	strcat((char *)optstr, "tr:");	// terminal mode and reset key
#endif
#ifdef X11
	strcat(optstr, "s");		// X11 synchronised
#endif
#if defined (X11) || defined(WIN32)
	strcat(optstr, "ic:");		// color, inverse video
#endif
#ifndef WIN32
	strcat((char *)optstr, "vh");	// version and help
#endif

	while (1) {
		result = getopt(argc, argv, optstr);
		if (result == -1)
			break;
		switch(result) {
			case 'f': pOptions->hex_file            = optarg; break;
			case '0': pOptions->drives[0]           = optarg; break;
			case '1': pOptions->drives[1]           = optarg; break;
			case '2': pOptions->drives[2]           = optarg; break;
			case '3': pOptions->drives[3]           = optarg; break;
			case 'p': pOptions->disk_dir            = optarg; break;
			case 'j': sscanf(optarg, "%d", &i);
				  if (i > 0 && i <= MAX_GUIXSIZE)
				  	pGuiOptions->guiXSize = i;
				  break;
			case 'k': sscanf(optarg, "%d", &i);
				  if (i > 0 && i <= MAX_GUIYSIZE)
				  	pGuiOptions->guiYSize = i;
				  break;
#ifdef UNIX
			case 't': pOptions->term_mode           = 1;      break;
			case 'r': sscanf(optarg, "%hx", &pOptions->reset_key);
				  break;
#endif
#ifdef X11
			case 's': pGuiOptions->synchronized     = 1;      break;
#endif
#if defined (X11) || defined(WIN32)
			case 'c': pGuiOptions->color            = optarg; break;
			case 'i': pGuiOptions->inverse          = 1;      break;
#endif
#ifndef WIN32
			case 'v':
				  fprintf(stdout, "flexemu V %s\n", VERSION);
				  exit(EXIT_SUCCESS);
			case '?':
			case 'h': print_help();
				  exit(EXIT_SUCCESS);
#endif
		}  // switch
	} // while
} // get_options


#ifdef WIN32
LONG create_value_win32(HKEY key, char *name, char *value)
{
	return RegSetValueEx(key, (LPCTSTR)name, 0, REG_SZ, (CONST BYTE *)value, strlen(value)+1);

} /* create_value_win32 */


void set_options_win32(void)
{
	HKEY key;
	int  i;
	char envName[16];
	char dir[PATH_MAX];

	if (RegOpenKeyEx(HKEY_CURRENT_USER, FLEXREG, 0,
		KEY_QUERY_VALUE, &key) != ERROR_SUCCESS) {
		if (RegCreateKeyEx(HKEY_CURRENT_USER, FLEXREG, 0, "Flexemu", REG_OPTION_NON_VOLATILE,
			KEY_ALL_ACCESS, NULL, &key, NULL) == ERROR_SUCCESS) {
			create_value_win32(key, FLEXINVERSE, "false");
			create_value_win32(key, FLEXCOLOR, "green");
			create_value_win32(key, FLEXMONITOR, "neumon54.hex");
			GetCurrentDirectory(PATH_MAX, dir);
			strcat(dir, "\\doc\\");
			create_value_win32(key, FLEXDOCDIR, dir);
			GetCurrentDirectory(PATH_MAX, dir);
			strcat(dir, "\\disks\\");
			create_value_win32(key, FLEXDISKDIR, dir);
			strcpy(envName, FLEXDISKX);
			if (strchr(envName, 'x'))
				*strchr(envName, 'x') = '0';
			create_value_win32(key, (char *)envName, "system.dsk");
			for (i = 1; i <= 3; i++) {
				strcpy(envName, FLEXDISKX);
				if (strchr(envName, 'x'))
					*strchr(envName, 'x') = i + '0';
				create_value_win32(key, (char *)envName, "");
			}
		}
		RegCloseKey(key);
	}
} /* set_options_win32 */

BOOL query_value_win32(HKEY key, char *name, LPBYTE pbuffer, DWORD maxSize)
{
	DWORD type, size;
	
	size = maxSize;
	return (RegQueryValueEx(key, (LPCTSTR)name, 0, &type, pbuffer, &size) == ERROR_SUCCESS);
} /* query_value_win32 */

void get_options_win32(struct sGuiOptions *pGuiOptions,
			struct sOptions *pOptions)
{
	int    i;
	char   envName[16];
	static char hex_file[PATH_MAX];
	static char disk_dir[PATH_MAX];
	static char doc_dir[PATH_MAX];
	static char disk[4][PATH_MAX];
	static char color[16];
	HKEY   key;

	if (RegOpenKeyEx(HKEY_CURRENT_USER, FLEXREG, 0,
	   KEY_QUERY_VALUE, &key) == ERROR_SUCCESS) {
		if (query_value_win32(key, FLEXINVERSE, (LPBYTE)&disk_dir, PATH_MAX))
			if (stricmp("true", disk_dir) == 0)
				pGuiOptions->inverse = 1;
			else
				pGuiOptions->inverse = 0;
		if (query_value_win32(key, FLEXDISKDIR, (LPBYTE)&disk_dir, PATH_MAX))
			pOptions->disk_dir = (char *)disk_dir;
		if (query_value_win32(key, FLEXMONITOR, (LPBYTE)&hex_file, PATH_MAX))
			pOptions->hex_file = (char *)hex_file;
		if (query_value_win32(key, FLEXCOLOR, (LPBYTE)&color, 16))
			pGuiOptions->color = (char *)color;
		if (query_value_win32(key, FLEXDOCDIR, (LPBYTE)&doc_dir, PATH_MAX))
			pGuiOptions->doc_dir = (char *)doc_dir;
		for (i = 0; i <= 3; i++) {
			strcpy(envName, FLEXDISKX);
			if (strchr(envName, 'x'))
				*strchr(envName, 'x') = i + '0';
			if (query_value_win32(key, envName, (LPBYTE)&disk[i], PATH_MAX))
				pOptions->drives[i] = (char *)disk[i];
		}
		RegCloseKey(key);
	}
} // get_options_win32
#endif WIN32

void startup(
	Mc6809             **cpu,	
	Inout              **io,
	Da6809             **disassembler,
	struct sGuiOptions *pGuiOptions,
	struct sOptions    *pOptions)
{
	E2video		*video;
	IoDevice	*device;
	Command		*comm;

	*cpu	          = new Mc6809();
	(*cpu)->initialize_io_page(GENIO_BASE);
	*disassembler	  = new Da6809(*cpu);
	*io		  = new Inout(*cpu, pGuiOptions);
	(*cpu)->set_disassembler(*disassembler);
	(*cpu)->set_inout(*io);

	// instanciate all memory mapped I/O devices

	device		  = new Mmu(*io, *cpu);
	PMEMORY->add_io_device(device, MMU_BASE, MMU_MASK, 0, 0);
	device		  = new Acia1(*io, *cpu);
	PMEMORY->add_io_device(device, ACIA1_BASE, ACIA1_MASK, 0, 0);
	device		  = new Pia1(*io, *cpu);
	PMEMORY->add_io_device(device, PIA1_BASE, PIA1_MASK, 0, 0);
	(*io)->set_pia1((Mc6821 *)device);
	device		  = new Pia2(*io, *cpu);
	PMEMORY->add_io_device(device, PIA2_BASE, PIA2_MASK, 0, 0);
	(*io)->set_pia2((Mc6821 *)device);
	E2floppy *fdc	  = new E2floppy(*io, *cpu);
	fdc->disk_directory(pOptions->disk_dir);
	fdc->mount_all_drives(pOptions->drives);
	PMEMORY->add_io_device(fdc, FDCA_BASE, FDCA_MASK,
				FDCB_BASE, FDCB_MASK);
	(*io)->set_fdc(fdc);
	comm		  = new Command(*io, *cpu);
	PMEMORY->add_io_device(comm, COMM_BASE, COMM_MASK, 0, 0);
	comm->set_fdc(fdc);
	video		  = new E2video(*io, *cpu);
	PMEMORY->add_io_device(video, VICO_BASE, VICO_MASK, 0, 0);
	(*io)->set_video(video);

	(*io)->init(pOptions->reset_key);
	if (!(pOptions->term_mode && (*io)->is_terminal_supported()))
#ifdef WIN32
		(*io)->create_gui(GUI_WINDOWS);
#else
#ifdef X11
#ifdef XTK
		(*io)->create_gui(GUI_XTOOLKIT);
#else
		(*io)->create_gui(GUI_X11);
#endif // XTK
#endif // X11
#endif // WIN32


	// instanciate real time clock right before initialize alarm
	device	 = new Mc146818(*io, *cpu);
	PMEMORY->add_io_device(device, RTC_LOW, RTC_HIGH-RTC_LOW+1, 0, 0);
	(*io)->set_rtc((Mc146818 *)device);

#ifndef WIN32
#ifdef SIGALRM
	(void)signal(SIGALRM, update);
	alarm(1);
#endif
#else
	idTimer = SetTimer(NULL, 0, 1000, CALLBACKCAST OneSecTimerProc);
#endif // WIN32

	(*cpu)->load_hexfile(pOptions->hex_file);

	// initialize monitor for parallel (0x00) or serial (0xff) output
	if (pGuiOptions->switch_sp &&
		pOptions->term_mode &&
		(*io)->is_terminal_supported())
		PMEMORY->write_rom(INITSP, 0xff);

	(*cpu)->reset_io();
	(*cpu)->reset();
} // startup

void shutdown (
	Mc6809      *cpu,	
	Inout       *io,
	Da6809      *disassembler)
{
#ifdef WIN32
	if (idTimer)
		KillTimer(NULL, idTimer);
#endif
	// gui is an instance of io
	if (io->gui != NULL)
		delete io->gui;
	delete io;
	delete disassembler;
	delete cpu;
} // shutdown

#ifndef WIN32
int main(int argc, char *argv[])
{
	struct		 sOptions    options;
	struct  	 sGuiOptions guiOptions;
	Mc6809		*cpu;
	Inout		*io;
	Da6809		*disassembler;

	init_options(&guiOptions, &options, argc, argv);
	get_options(&guiOptions, &options, argc, argv);

	startup(&cpu, &io, &disassembler, &guiOptions, &options);

	cpu->statemachine1();

	shutdown(cpu, io, disassembler);

	exit(0);
	return 0; // satisfy compiler
}

#else

void scanCmdLine(LPSTR lpCmdLine, int *argc, char **argv) {
	*argc = 1;
	*(argv + 0) = "flexemu";
	while (*lpCmdLine) {
		*(argv + *argc) = lpCmdLine;
		while (*lpCmdLine && *lpCmdLine != ' ' && *lpCmdLine != '\t')
			lpCmdLine++;
		if (*lpCmdLine)
			*(lpCmdLine++) = '\0';
		while (*lpCmdLine && (*lpCmdLine == ' ' || *lpCmdLine == '\t'))
			lpCmdLine++;
		(*argc)++;
	}
} // scanCmdLine

int WINAPI WinMain (
	HINSTANCE hInstance,
	HINSTANCE hPrevInstance,
	LPSTR lpCmdLine,
	int nCmdShow)
{
	struct		 sOptions    options;
	struct  	 sGuiOptions guiOptions;
	Mc6809		*cpu;
	Inout		*io;
	Da6809		*disassembler;
	int		argc;
	char		*argv[50];

	guiOptions.hInstance = hInstance;
	guiOptions.nCmdShow  = nCmdShow;
	set_options_win32();
	scanCmdLine(lpCmdLine, &argc, (char **)argv);
	init_options(&guiOptions, &options, argc, argv);
	get_options_win32(&guiOptions, &options);
	get_options(&guiOptions, &options, argc, argv);
	startup(&cpu, &io, &disassembler, &guiOptions, &options);

	cpu->statemachine1();

	shutdown(cpu, io, disassembler);

	return 0;
} // WinMain

#endif // #ifdef WIN32

