/*
    nafs.cc


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "misc1.h"

// support of the native file system

#ifdef NAFS
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef WIN32
#include <windows.h>
#include <sys/utime.h>
#ifdef _MSC_VER
#include <io.h>			// needed for access()
#endif
#else
#include <dirent.h>
#include <unistd.h>
#include <utime.h>
#endif
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#ifndef PATH_MAX
#ifdef WIN32
#define PATH_MAX MAX_PATH
#else
#include <sys/param.h>
#endif
#endif
#include "e2floppy.h"


// detailed debug messegas can be written to a debug file:
// #define DEBUG_FILE "nafs.log"

#ifdef DEBUG_FILE
#include "debug.h"
#endif


void E2floppy::header_for_nafs_format(s_floppy *pfloppy, Byte wp) {
	pfloppy->ready         = 1;
	pfloppy->type          = NAFS_FORMAT;
	pfloppy->offset        = 0;
	pfloppy->fp            = NULL;
	pfloppy->write_protect = wp;
	pfloppy->max_sector    = MAX_SECTOR >> 1;
	pfloppy->max_sector0   = (INIT_DIR_SECTORS + 4) >> 1;
	pfloppy->max_track     = MAX_TRACK;
	pfloppy->byte_p_sector = SECTOR_SIZE;
	pfloppy->byte_p_track0 = pfloppy->max_sector0 << 9; 
	pfloppy->byte_p_track  = pfloppy->max_sector << 9; 
	pfloppy->pflex_links     = (struct s_link_table *)calloc(LINK_TABLE_SIZE, sizeof(struct s_link_table));
	pfloppy->pflex_sys_info  = (struct s_sys_info_sector *)calloc(2, sizeof(s_sys_info_sector));
	pfloppy->pflex_unused    = (struct s_unused_sector *)calloc(1, sizeof(s_unused_sector));
	pfloppy->pflex_directory = (struct s_dir_sector *)calloc(INIT_DIR_SECTORS, sizeof(struct s_dir_sector));
	pfloppy->dir_sectors	 = INIT_DIR_SECTORS;
	pfloppy->pnew_file       = (struct s_new_file *)calloc(INIT_NEW_FILES, sizeof(s_new_file));
	pfloppy->new_files	 = INIT_NEW_FILES;
	pfloppy->dir_extend.sec_trk = 0;
	if (pfloppy->pflex_links == NULL || pfloppy->pflex_sys_info == NULL ||
	    pfloppy->pflex_unused == NULL || pfloppy->pflex_directory == NULL ||
	    pfloppy->pnew_file == NULL) {
		char *pmsg = "Not enough memory, can't continue\n";
#ifdef WIN32
		MessageBox(NULL, pmsg, PROGRAMNAME " error",
			MB_OK | MB_ICONERROR);
#else
		fprintf(stderr, pmsg);
#endif
		exit(EXIT_FAILURE);
	} // if
} // header_for_nafs_format

// if valid flex filename return is > 0 otherwise = 0
/*
	some examples:

	allowed:	x.a xx.a xxxxxxxx.a x xx xxxxxxxx
	not allowed:	x. .a xx. xxxxxxxxx.a X.a xxxxxxxxX.a
*/
Byte E2floppy::is_flex_filename(char *pfilename, char *pname, char *pext)
{
	int	result; // result from sscanf should be int
	char	dot;

	*pname = '\0';
	*pext  = '\0';
	dot    = '\0';
	result = sscanf(pfilename, "%1[a-z]%7[a-z0-9_-]", pname, (pname + 1));
	if (!result || result == EOF)
		return 0;
	if (result == 1)
		result = sscanf(pfilename, "%*1[a-z]%c%1[a-z]%2[a-z0-9_-]",
		&dot, pext, (pext + 1));
	else
		result = sscanf(pfilename, "%*1[a-z]%*7[a-z0-9_-]%c%1[a-z]%2[a-z0-9_-]",
		&dot, pext, (pext + 1));
	if (!result || result == 1 || result == EOF)
		return 0;
	if (strlen(pname) + strlen(pext) + (dot == '.' ? 1 : 0) != strlen(pfilename))
		return 0;
	strupper(pname);
	strupper(pext);
	return 1;
} // is_flex_filename


void E2floppy::initialize_flex_directory(struct s_floppy *pfloppy)
{
	Word i, j;
	char *p;

	for (i = 0; i < pfloppy->dir_sectors; i++) {
		p = (char *)(pfloppy->pflex_directory + i);
		for (j=0; j < SECTOR_SIZE; j++)
			*(p++) = 0x00;
		if (i < pfloppy->dir_sectors - 1) {
			(pfloppy->pflex_directory+i)->next_trk = (i+5) / MAX_SECTOR;
			(pfloppy->pflex_directory+i)->next_sec =
				((i+5) % MAX_SECTOR) + 1;
		} // if
	} // for
} // initialize_flex_directory


void E2floppy::scan_unix_filename(char *pfn, char *pfilename)
{
	char *p;
	Word i, j;

	j = 0;
	p = (char *)pfn;
	if (*p != -1) {
		// get file name
		for (i=0; i < 8 && *p; i++) {
			*(pfilename+i) = tolower(*p);
			p++;
		}
		if (i > 0) {
			p += 8 - i;
			*(pfilename + (i++)) = '.';
			// get file extension
			for (j=i; j < i+3 && *p; j++) {
				*(pfilename+j) = tolower(*p);
				p++;
			} // for
		} // if
	} // if
	*(pfilename+j) = '\0';
} // scan_unix_filename


// return unix filename from a pointer to a flex directory entry 
char *E2floppy::to_unix_filename(s_dir_entry *pde)
{
	static char filename[12];

	scan_unix_filename((char *)pde, (char *)&filename);
	return (char *)&filename;
} // to_unix_filename


// return unix filename from a given file_id
char *E2floppy::unix_filename(SWord file_id)
{
	static char filename[12];

	if (file_id < 0)
		sprintf((char *)&filename, "tmp%02d",  NEW_FILE1 - file_id);
	else
		scan_unix_filename(
		   (char *)&((pfs->pflex_directory + (file_id/10))->dir_entry[file_id%10].filename),
		   (char *)&filename);
	return (char *)&filename;
} // unix_filename

// return the record nr (beginning from 0) of a new file which first
// sector has index 'index' into flex link table
Word E2floppy::record_nr_of_new_file(SWord new_file_index, Word index)
{
	SWord 	i;
	Word	record_nr;

	record_nr = 0;
	i = (pfs->pnew_file + new_file_index)->first.st.trk * MAX_SECTOR +
		(pfs->pnew_file + new_file_index)->first.st.sec - 1;
	while (i != index && i >= 0) {
		record_nr++;
		i = (pfs->pflex_links+i)->next.st.trk * MAX_SECTOR +
			(pfs->pflex_links+i)->next.st.sec - 1;
	} // while
	return record_nr;
}

SWord E2floppy::index_of_new_file(Byte track, Byte sector)
{
	SWord		i;
	char		path[PATH_MAX+1];
	s_new_file	*pnf;

	i = 0;
	while ((pfs->pnew_file+i)->first.sec_trk && i < pfs->new_files) {
		if (tr == (pfs->pnew_file+i)->next.st.trk &&
		    sr == (pfs->pnew_file+i)->next.st.sec)
			return i;
		i++;
	} // while
	if (i >= pfs->new_files) {
		// new file table is full, increase new file table by 2
		pnf = (s_new_file *)calloc(pfs->new_files + 2,
			sizeof(struct s_new_file));
		if (pnf == NULL)
			return -1; // can't allocate memory
		memcpy(pnf, pfs->pnew_file,
			pfs->new_files * sizeof(struct s_new_file));
		free(pfs->pnew_file);
		pfs->pnew_file = pnf;
		pfs->new_files += 2;
		(pfs->pnew_file+(pfs->new_files-2))->first.sec_trk = 0;
		(pfs->pnew_file+(pfs->new_files-1))->first.sec_trk = 0;
	} // if
	strcpy((char *)&(pfs->pnew_file+i)->filename,
		unix_filename(NEW_FILE1 - i));
	strcpy((char *)&path, pfs->ppath);
	strcat((char *)&path, PATHSEPARATORSTRING);
	strcat((char *)&path, (pfs->pnew_file+i)->filename);
#ifdef sun
	(pfs->pnew_file+i)->fp = fopen(path, "w+");
#else
	(pfs->pnew_file+i)->fp = fopen(path, "wb+");
#endif
	if ((pfs->pnew_file+i)->fp == NULL)
		return -1;
	(pfs->pnew_file+i)->first.st.trk = track;
	(pfs->pnew_file+i)->first.st.sec = sector;
	(pfs->pnew_file+i)->next.sec_trk = 0;
	(pfs->pnew_file+i)->f_record     = 0;
	return i;
} // index_of_new_file


Byte E2floppy::extend_directory(
	struct s_floppy	*pfloppy,
	SWord		index,
	s_dir_sector	*pdb)
{
	struct s_dir_sector *pfd;

	// increase flex_directory by one sector
	pfd = (s_dir_sector *)calloc(pfloppy->dir_sectors + 1,
		sizeof(struct s_dir_sector));
	if (pfd == NULL)
		return 0xff; // can't allocate memory
	memcpy(pfd, pfloppy->pflex_directory,
		pfloppy->dir_sectors * sizeof(struct s_dir_sector));
	memcpy(pfd+pfloppy->dir_sectors, pdb, SECTOR_SIZE);
	(pfloppy->pflex_links+index)->f_record = pfloppy->dir_sectors;
	(pfloppy->pflex_links+index)->file_id  = DIRECTORY;
	free(pfloppy->pflex_directory);
	pfloppy->pflex_directory = pfd;
	pfloppy->dir_sectors++;
	pfloppy->dir_extend.sec_trk = 0;// reset directory extend track/sector
	return 0;
} // extend_directory


t_st *E2floppy::link_address(void)
{
	static t_st	link;
	s_dir_entry	*pd;
	Word		i;

	link.sec_trk = 0;
	for (i = 0; i < pfs->dir_sectors * 10; i++) {
		pd = &((pfs->pflex_directory + (i / 10))->dir_entry[i % 10]);
		if (!strncmp(pd->filename, "FLEX\0\0\0\0", 8) &&
			!strncmp(pd->file_ext, "SYS", 3)) {
			link.st.trk = pd->start_trk;		
			link.st.sec = pd->start_sec;
			break;
		}		
	} // for
	return &link;
}  // link_address


//if directory can't be extended return -1
SWord E2floppy::next_free_dir_entry(struct s_floppy *pfloppy)
{
	SWord			i;
	Word			j, index;
	s_sys_info_sector	*psis;
	char			sector_buffer[SECTOR_SIZE];
	Word			trk, sec;

	for (i = 0; i < pfloppy->dir_sectors * 10; i++) {
		if (!((pfloppy->pflex_directory + (i / 10))->dir_entry[i % 10].filename[0]))
			return i;
	} // for
	psis = pfloppy->pflex_sys_info;
	for (j = 0; j < SECTOR_SIZE; j++)
		sector_buffer[j] = 0x00;
	sector_buffer[3] = pfloppy->dir_sectors - INIT_DIR_SECTORS + 1;
	trk = psis->fc_start_trk;
	sec = psis->fc_start_sec;
	index = trk * MAX_SECTOR + sec - 1;
	if (!extend_directory(pfloppy, index, (s_dir_sector *)sector_buffer)) {
		(pfloppy->pflex_directory+pfloppy->dir_sectors-2)->next_trk =
			(Byte)trk;
		(pfloppy->pflex_directory+pfloppy->dir_sectors-2)->next_sec =
			(Byte)sec;
		if (++psis->fc_start_sec > MAX_SECTOR) {
			psis->fc_start_sec = 1;
			psis->fc_start_trk++;
		}
		if (--psis->free[1] == 0xff)
			psis->free[0]--;
		return i;
	} else
		return -1;
}  // next_free_dir_entry


void E2floppy::initialize_flex_link_table(struct s_floppy *pfloppy)
{
	Word				 i, fc_start, free;
	struct s_link_table		*plink;
	struct s_sys_info_sector	*psis;

	fc_start = MAX_SECTOR;
	// entries for system and directory sectors
	for (i = 0; i < fc_start; i++) {
		plink = pfloppy->pflex_links + i;
		plink->next.sec_trk	= 0;
		plink->record_nr[0]	= 0;
		plink->record_nr[1]	= 0;
		plink->f_record		= i < 4 ? 0 : i - 4;
		plink->file_id		= i < 4 ? SYSTEM : DIRECTORY;
	} // for
	for (i = fc_start; i < LINK_TABLE_SIZE; i++) {
		plink = pfloppy->pflex_links + i;
		if (i == LINK_TABLE_SIZE-1)
			plink->next.sec_trk = 0;
		else {
			plink->next.st.trk = (i+1) / MAX_SECTOR;
			plink->next.st.sec = ((i+1) % MAX_SECTOR) + 1;
		} // else
		plink->record_nr[0]	= 0;
		plink->record_nr[1]	= 0;
		plink->f_record		= 0;
		plink->file_id		= FREE_CHAIN;
	} // for
	free = LINK_TABLE_SIZE - fc_start;
	// and now update system info sectors
	for (i = 0; i < 2; i++) {
		psis = pfloppy->pflex_sys_info + i;
		psis->fc_start_trk	= fc_start / MAX_SECTOR;
		psis->fc_start_sec	= (fc_start % MAX_SECTOR) + 1;
		psis->fc_end_trk	= MAX_TRACK;
		psis->fc_end_sec	= MAX_SECTOR;
		psis->free[0]		= free >> 8;
		psis->free[1] 		= free & 0xff;
	} // for
} // initialize_flex_link_table


void E2floppy::initialize_new_file_table(struct s_floppy *pfloppy)
{
	Word i;

	for (i = 0; i < pfloppy->new_files; i++) {
		(pfloppy->pnew_file+i)->first.sec_trk = 0;
		(pfloppy->pnew_file+i)->next.sec_trk = 0;
	} // for
} // initialize_new_file_table

// check for any open file in an nafs file system
Byte E2floppy::open_files(Word index)
{
	struct s_floppy	*pfloppy;
	Word		i;

	pfloppy = &floppy[index];
	if (pfloppy->type == NAFS_FORMAT) {
		for (i = 0; i < pfloppy->new_files; i++) {
			if ((pfloppy->pnew_file+i)->first.sec_trk != 0)
				return 1;
		} // for
	} // if
	return 0;
} // open_files


void E2floppy::close_new_files(Word index)
{
	struct s_floppy	*pfloppy;
	Word		i;
	Byte		first = 1;
	char		msg[ERR_SIZE];

	msg[0] = '\0';
	pfloppy = &floppy[index];
	if (pfloppy->type == NAFS_FORMAT) {
		for (i = 0; i < pfloppy->new_files; i++) {
			if ((pfloppy->pnew_file+i)->first.sec_trk != 0) {
				if (first) { 
					strcpy(msg, "There are still open files\ntemporarily stored as:\n");
					first = 0;
				}
				fclose((pfloppy->pnew_file+i)->fp);
				if (strlen(msg) < (ERR_SIZE - 18)) {
					strcat(msg, "   ");
					strcat(msg, (pfloppy->pnew_file+i)->filename);
					strcat(msg, "\n");
				} // if
			} // if
		} // for
		if (strlen(msg) > 0)
#ifdef WIN32
			MessageBox(NULL, msg, PROGRAMNAME " warning",
			MB_OK | MB_ICONEXCLAMATION);
#else
			fprintf(stderr, msg);
#endif
	} // if
} // close_new_files


void E2floppy::free_nafs_memory(Word index)
{
	if ((floppy[index].type == NAFS_FORMAT)) {
		free(floppy[index].pflex_links);
		free(floppy[index].pflex_directory);
		free(floppy[index].pflex_sys_info);
		free(floppy[index].pflex_unused);
		free(floppy[index].pnew_file);
	} // if
} // free_nafs_memory

// if file won't fit return 0 otherwise return 1
Byte E2floppy::add_to_link_table(
	struct s_floppy	*pfloppy,
	SWord		dir_index,
	off_t		size,
	Byte		random,
	t_st		*pbegin,
	t_st		*pend)
{
	Word				i, free, begin, records;
	struct s_link_table		*plink;
	struct s_sys_info_sector	*psis;

	psis = pfloppy->pflex_sys_info;	
	free = (psis->free[0] << 8) + psis->free[1];
	if (size > (off_t)(free * 252L))
		return 0;
	records = (size + 251L) / 252;
	pbegin->st.sec = psis->fc_start_sec;
	pbegin->st.trk = psis->fc_start_trk;
	begin = (pbegin->st.trk * MAX_SECTOR) + pbegin->st.sec - 1;
	for (i = 1; i <= records; i++) {
		plink = pfloppy->pflex_links + i + begin - 1;
		if (i == records) 
			plink->next.sec_trk = 0;
		if (random) {
			plink->record_nr[0]	= i > 2 ? (i-2) >> 8 : 0;
			plink->record_nr[1]	= i > 2 ? (i-2) & 0xff : 0;
		} else {
			plink->record_nr[0]	= i >> 8;
			plink->record_nr[1]	= i & 0xff;
		}
		plink->f_record		= i-1;
		plink->file_id		= dir_index;
	} // for
	pend->st.sec = ((i + begin - 2) % MAX_SECTOR) + 1;
	pend->st.trk = (i + begin - 2) / MAX_SECTOR;
	// update sys info sector
	psis->fc_start_sec = ((i + begin - 1) % MAX_SECTOR) + 1;
	psis->fc_start_trk = (i + begin -1) / MAX_SECTOR;
	psis->free[0] = (free - records) >> 8;
	psis->free[1] = (free - records) & 0xff;
	return 1;
} // add_to_link_table


void E2floppy::add_to_directory(
	struct s_floppy	*pfloppy,
	char		*name,
	char		*ext,
	SWord		dir_index,
	Byte		random,
	struct		stat *pstat,
	t_st		*pbegin,
	t_st		*pend,
	Byte		wp)
{
	s_dir_entry	*pd;
	struct tm	*lt;
	SWord		records;

	lt = localtime(&(pstat->st_mtime));
	records = (pstat->st_size + 251L) / 252;
	pd = &((pfloppy->pflex_directory + (dir_index / 10))->dir_entry[dir_index % 10]);
	strcpy(pd->filename, name);
	strcpy(pd->file_ext, ext);
	pd->file_attr	= wp ? 0xc0 : 0x00;
	pd->start_trk	= pbegin->st.trk;
	pd->start_sec	= pbegin->st.sec;
	pd->end_trk	= pend->st.trk;
	pd->end_sec	= pend->st.sec;
	pd->records[0]	= records >> 8;
	pd->records[1]	= records & 0xff;
	pd->sector_map	= random ? 0x02 : 0x00;
	pd->month	= lt->tm_mon + 1;
	pd->day		= lt->tm_mday;
	pd->year	= lt->tm_year;
} // add_to_directory

Word E2floppy::is_in_file_random(const char *ppath, const char *pfilename)
{
	FILE	*fp;
	char	str[PATH_MAX+1];

	strcpy(str, ppath);
	strcat(str, PATHSEPARATORSTRING "random");
	if ((fp = fopen(str, "r")) != NULL) {
		while (!feof(fp)) {
			fgets(str, PATH_MAX, fp);
			if (strchr(str, '\n'))
				*strchr(str, '\n') = '\0';
			if (strcmp(pfilename, str) == 0)
				return 1;
		}
		fclose(fp);
	} // if
	return 0;

} // is_in_file_random

void E2floppy::modify_random_file(char *path, struct stat *pstat, t_st *pbegin)
{
	Byte	file_sector_map[252 * 2];
	SDWord	data_size;
	Word	i, n, index;
	FILE	*fp;

	data_size = pstat->st_size - (252L * 2);
	if (data_size >= 252L * 2) {
		index = (pbegin->st.trk * MAX_SECTOR + pbegin->st.sec) - 1 + 2;
		for (i = 0; i < 252 * 2; i++)
			file_sector_map[i] = 0;
		n = 0;
		for (n = 0; n < (data_size / (252L * 255)) ; n++) {
			file_sector_map[3*n]   = index / MAX_SECTOR;
			file_sector_map[3*n+1] = (index % MAX_SECTOR) + 1;
			file_sector_map[3*n+2] = 255;
			index += 255;
		} // for
		i = (Word)(data_size % (252L * 255));
		if (i != 0) {
			file_sector_map[3*n]   = index / MAX_SECTOR;
			file_sector_map[3*n+1] = (index % MAX_SECTOR) + 1;
			file_sector_map[3*n+2] = (i+251) / 252;
		} // if
#ifdef sun
		fp = fopen(path, "r+");
#else
		fp = fopen(path, "rb+");
#endif
		if (fp != NULL)
			fwrite(&file_sector_map, 252, 2, fp);
		fclose(fp);
	} // if
} // modify_random_file

// dwp = write protect for drive

void E2floppy::fill_flex_directory(struct s_floppy *pfloppy, Byte dwp)
{
#ifdef WIN32
	HANDLE		hdl;
	WIN32_FIND_DATA	pentry;
#else
	DIR		*pd;
	struct dirent	*pentry;
#endif
	char		name[9], ext[4], path[PATH_MAX+1], *pfilename;
	SWord		dir_index;
	Word		random;
	t_st		begin, end;
	struct stat	sbuf;

	initialize_flex_directory(pfloppy);
	initialize_flex_link_table(pfloppy);
	dir_index = 0;
#ifdef WIN32
	strcpy(path, pfloppy->ppath);
	strcat(path, PATHSEPARATORSTRING "*.*");
	if ((hdl = FindFirstFile(path, &pentry)) != INVALID_HANDLE_VALUE) {
		do {
#else
	if ((pd = opendir(pfloppy->ppath)) != NULL) {
		while ((pentry = readdir(pd)) != NULL) {
#endif
#ifdef WIN32
			pfilename = pentry.cFileName;
#else
			pfilename = (char *)&pentry->d_name;
#endif
			random = 0;
			strcpy((char *)&path, pfloppy->ppath);
			strcat((char *)&path, PATHSEPARATORSTRING);
			strcat((char *)&path, pfilename);
			if (is_flex_filename(pfilename, (char *)&name, (char *)&ext) &&
			   !stat(path, &sbuf) && (sbuf.st_mode & S_IFREG) &&
			   sbuf.st_size > 0 &&
			   (dir_index = next_free_dir_entry(pfloppy)) >= 0) {
#ifdef WIN32
				random = (pentry.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) ? 1 : 0;
#else
			   	random = sbuf.st_mode & S_IXUSR;
#endif
				// CDFS-Support: look for file name in file 'random'
				if (dwp)
					random |= is_in_file_random(pfloppy->ppath, pfilename);
				if (add_to_link_table(pfloppy, dir_index,
			   		sbuf.st_size, (Byte)random, &begin, &end)) {
			   		add_to_directory(pfloppy, name, ext,
			   	 	   dir_index, (Byte)random, &sbuf,
					   &begin, &end, dwp || access(path, W_OK));
			   		if (random)
			   			modify_random_file(path,
			   				&sbuf, &begin);
				}
			}
#ifdef WIN32
		} while (FindNextFile(hdl, &pentry) != 0);
		FindClose(hdl);
#else
		} // while
		closedir(pd);
#endif
	} //if 
} // initialize_flex_directory


void E2floppy::initialize_flex_sys_info_sectors(
	struct s_floppy	*pfloppy,
	Word		number)
{
	Word				i;
	char				name[9], ext[4], *pName;
	struct stat			sbuf;
	struct tm			*lt;
	struct s_sys_info_sector	*psis;

	if(!stat(pfloppy->ppath, &sbuf)) {
		psis = pfloppy->pflex_sys_info;
		lt = localtime(&(sbuf.st_mtime));
		name[0] = '\0';
		ext[0]  = '\0';
		pName = pfloppy->ppath + strlen(pfloppy->ppath) - 1;
		if (*pName == PATHSEPARATOR)
			pName--;
		while (pName != pfloppy->ppath && *pName != PATHSEPARATOR)
			pName--;
		if (*pName == PATHSEPARATOR)
			pName++;
		if (sscanf(pName, "%8[a-zA-Z0-9_-]", name) != 1)
			strcpy(name, "UNIX");
		else {
			strupper(name);
			strupper(ext);
		} // else
		for (i=0; i < 16; i++)
			psis->unused1[i] = 0;
		strncpy((char *)&(psis->disk_name), name, 8);
		strncpy((char *)&(psis->disk_ext), ext, 3);
		psis->disk_number[0] = number >> 8;
		psis->disk_number[1] = number & 0xff;
		psis->fc_start_trk   = 0;
		psis->fc_start_sec   = 0;
		psis->fc_end_trk     = 0;
		psis->fc_end_sec     = 0;
		psis->free[0]        = 0;
		psis->free[1]        = 0;
		psis->month          = lt->tm_mon + 1;
		psis->day            = lt->tm_mday;
		psis->year           = lt->tm_year;
		psis->last_trk       = MAX_TRACK;
		psis->last_sec       = MAX_SECTOR;
	
		for (i=0; i < 216; i++)
			psis->unused2[i] = 0;
		memcpy(pfloppy->pflex_sys_info + 1, pfloppy->pflex_sys_info,
			sizeof(struct s_sys_info_sector));
	} // if
} // initialize_flex_sys_info_sectors


// I don't know what this sector should be used for but
// emulate it anyway
void E2floppy::initialize_flex_unused_sector(struct s_floppy *pfloppy)
{
	Word i;

	pfloppy->pflex_unused->next_trk = 0x00;
	pfloppy->pflex_unused->next_sec = 0x03;
	for (i=0; i < 254; i++)
		pfloppy->pflex_unused->unused[i] = 0;
} // initialize_flex_unused_sector


// change the file id in flex_links
void E2floppy::change_file_id(SWord index, SWord old_id, SWord new_id)
{
	SWord i;

	i = index;
	while (i >= 0 && (pfs->pflex_links+i)->file_id == old_id) {
		(pfs->pflex_links+i)->file_id = new_id;
		i = (pfs->pflex_links+i)->next.st.trk * MAX_SECTOR +
			(pfs->pflex_links+i)->next.st.sec - 1;
	} // while
} // change_file_id


void E2floppy::check_for_delete(SWord dir_index, s_dir_sector *pdb)
{
	Word		i;
	SWord		index;
	char		path[PATH_MAX+1], *pfilename;
	s_dir_sector	*pd;

	pd = pfs->pflex_directory + dir_index;
	for (i = 0; i < 10; i++) {
		if (pdb->dir_entry[i].filename[0] == -1 &&
		   pd->dir_entry[i].filename[0] != -1) {
			pfilename = unix_filename(dir_index * 10 + i);
			index = pd->dir_entry[i].start_trk * MAX_SECTOR +
				pd->dir_entry[i].start_sec - 1;
			strcpy((char *)&path, pfs->ppath);
			strcat((char *)&path, PATHSEPARATORSTRING);
			strcat((char *)&path, pfilename);
			unlink(path);
			change_file_id(index, dir_index * 10 + i, FREE_CHAIN);
#ifdef DEBUG_FILE
			LOG_X("     delete %s\n", pfilename);
#endif
			break;
		} // if
	} // for	

} // check_for_delete


void E2floppy::check_for_rename(SWord dir_index, s_dir_sector *pdb)
{
	Word	i;
	char	old_path[PATH_MAX+1], new_path[PATH_MAX+1];
	char	*pfilename1, *pfilename2;

	for (i = 0; i < 10; i++) {
		pfilename1 = unix_filename(dir_index * 10 + i);
		pfilename2 = to_unix_filename(&pdb->dir_entry[i]);
		if (*pfilename1 && *pfilename2 &&
		   strcmp(pfilename1, pfilename2) != 0) {
			strcpy((char *)&old_path, pfs->ppath);
			strcat((char *)&old_path, PATHSEPARATORSTRING);
			strcat((char *)&old_path, pfilename1);
			strcpy((char *)&new_path, pfs->ppath);
			strcat((char *)&new_path, PATHSEPARATORSTRING);
			strcat((char *)&new_path, pfilename2);
			rename(old_path, new_path);
#ifdef DEBUG_FILE
			LOG_XX("     rename %s to %s\n",
				pfilename1, pfilename2);
#endif
			break;
		} // if
	} // for	
} // check_for_rename


void E2floppy::check_for_extend(SWord dir_index, s_dir_sector *pdb)
{
	s_dir_sector *pd;

	pd = pfs->pflex_directory + dir_index;
	if (!pd->next_sec && !pd->next_trk && (pdb->next_trk || pdb->next_sec)) {
		   pfs->dir_extend.st.trk = pdb->next_trk;
		   pfs->dir_extend.st.sec = pdb->next_sec;
	} // if
} // check_for_extend


// return -1 if not successful otherwise return 0
// years < 70 will be represented as >= 2000
SWord E2floppy::set_file_time(char *ppath, Byte month, Byte day, Byte year)
{
	struct stat    statbuf;
	struct utimbuf timebuf;
	struct tm      file_time;

	if (stat(ppath, &statbuf) >= 0) {
		timebuf.actime = statbuf.st_atime;
		file_time.tm_sec   = 0;
		file_time.tm_min   = 0;
		file_time.tm_hour  = 0;
		file_time.tm_mon   = month-1;
		file_time.tm_mday  = day;
		file_time.tm_year  = year < 70 ? year + 100 : year;
		file_time.tm_isdst = 0;
		timebuf.modtime    = mktime(&file_time);
		if (timebuf.modtime >= 0 && utime(ppath, &timebuf) >= 0)
				return 0;
			else
				return -1;
	} // if
	return -1;
} // set_file_time


Byte E2floppy::check_for_new_file(SWord dir_index, s_dir_sector *pd)
{
	Word		i, j;
	SWord		index;
	char		old_path[PATH_MAX+1], new_path[PATH_MAX+1];
#ifndef WIN32
	struct stat	sbuf;
#endif

	j = 0;
	while ((pfs->pnew_file+j)->first.sec_trk && j < pfs->new_files) {
		for (i = 0; i < 10; i++) {
			if ((pfs->pnew_file+j)->first.st.sec == pd->dir_entry[i].start_sec &&
			    (pfs->pnew_file+j)->first.st.trk == pd->dir_entry[i].start_trk) {
				index = (pfs->pnew_file+j)->first.st.trk * MAX_SECTOR +
			   	   (pfs->pnew_file+j)->first.st.sec - 1; 
				change_file_id(index, NEW_FILE1-j,
					10 * dir_index + i);
				fclose((pfs->pnew_file+j)->fp);
				strcpy((char *)&old_path, pfs->ppath);
				strcat((char *)&old_path, PATHSEPARATORSTRING);
				strcat((char *)&old_path, (pfs->pnew_file+j)->filename);
				// check for random file, if true set user execute bit
				if (pd->dir_entry[i].sector_map & 0x02)
#ifdef WIN32
					SetFileAttributes(old_path, FILE_ATTRIBUTE_HIDDEN);
#else
					if (!stat(old_path, &sbuf))
						chmod(old_path, sbuf.st_mode | S_IXUSR);
#endif
				strcpy((char *)&new_path, pfs->ppath);
				strcat((char *)&new_path, PATHSEPARATORSTRING);
				strcat((char *)&new_path,
					unix_filename((pfs->pflex_links+index)->file_id));
				rename(old_path, new_path); 
#ifdef DEBUG_FILE
			LOG_XX("     new file %s, was %s\n",
				unix_filename((pfs->pflex_links+index)->file_id),
				(pfs->pnew_file+j)->filename);
#endif
				set_file_time(new_path, pd->dir_entry[i].month,
				   pd->dir_entry[i].day, pd->dir_entry[i].year);
				// remove entry in new file table and
				// move following entries
				while ((pfs->pnew_file+j+1)->first.sec_trk &&
					j < pfs->new_files-1) {
					strcpy((char *)&(pfs->pnew_file+j)->filename,
						(pfs->pnew_file+j+1)->filename);
					(pfs->pnew_file+j)->first.sec_trk = (pfs->pnew_file+j+1)->first.sec_trk;
					(pfs->pnew_file+j)->next.sec_trk  = (pfs->pnew_file+j+1)->next.sec_trk;
					(pfs->pnew_file+j)->f_record      = (pfs->pnew_file+j+1)->f_record;
					(pfs->pnew_file+j)->fp            = (pfs->pnew_file+j+1)->fp;
					index = (pfs->pnew_file+j)->first.st.trk * MAX_SECTOR +
			   		   (pfs->pnew_file+j)->first.st.sec - 1; 
					change_file_id(index, NEW_FILE1-j-1, NEW_FILE1-j);
					j++;
				} // while
				(pfs->pnew_file+j)->first.sec_trk = 0;
				return 0;
			} // if
		} // for
		j++;
	} // while
	return 0;
} // check_for_new_file


Byte E2floppy::last_of_free_chain(Byte tr, Byte sr)
{
	return (pfs->pflex_sys_info->fc_end_trk == tr &&	
		pfs->pflex_sys_info->fc_end_sec == sr);
} // last_of_free_chain


void E2floppy::read_nafs_sector(void)
{
	char			*p, path[PATH_MAX+1], *mode;
	Word			i, di;
	size_t			bytes;
	FILE			*fp;
	t_st			*link;
	struct s_link_table	*pfl;

#ifdef DEBUG_FILE
	LOG_XX("read: %02X/%02X ", tr, sr);
#endif
#ifdef sun
	mode = "r";
#else
	mode = "rb";
#endif
	i = tr * MAX_SECTOR + (sr - 1);
	pfl = pfs->pflex_links+i;
	switch (pfl->file_id) {
		case SYSTEM :
#ifdef DEBUG_FILE
			LOG("systemsector\n");
#endif
			p = NULL; // satisfy compiler
			if (sr >= 3)
				p = (char *)(pfs->pflex_sys_info + sr - 3);
			else if (sr == 2)
				p = (char *)pfs->pflex_unused;
			else if (sr == 1) {
				bytes = SECTOR_SIZE;
				for (i = 0; i < SECTOR_SIZE; i++)
					sector_buffer[i] = 0;
				sector_buffer[0] = 0x39; // means RTS
				link = link_address();
				strcpy((char *)&path, pfs->ppath);
				strcat((char *)&path, PATHSEPARATORSTRING "boot");
				if ((fp = fopen(path, mode)) != NULL) {
					bytes = fread(&sector_buffer, 1,
						SECTOR_SIZE, fp);
					fclose(fp);
				}
				if (bytes != SECTOR_SIZE) {
					drq = 0;
					str = 0x10;
					byteCount = 0;
				}
				sector_buffer[3] = link->st.trk;
				sector_buffer[4] = link->st.sec;
				break;
			} else
				break;
			memcpy(sector_buffer, p, pfs->byte_p_sector);
			break;
		case DIRECTORY:
#ifdef DEBUG_FILE
			LOG("directorysector\n");
#endif
			di = pfl->f_record;
			p = (char *)(pfs->pflex_directory + di);
			memcpy(&sector_buffer, p, pfs->byte_p_sector);
			break;
		case FREE_CHAIN :
#ifdef DEBUG_FILE
			LOG("free chain\n");
#endif
			sector_buffer[0] = pfl->next.st.trk;
			sector_buffer[1] = pfl->next.st.sec;
			sector_buffer[2] = pfl->record_nr[0];
			sector_buffer[3] = pfl->record_nr[1];
			// free chain sector reads always
			// filled with zero
			for (i = 4; i < SECTOR_SIZE; i++)
				sector_buffer[i] = 0;
			break;
		default : 
			if (pfl->file_id >= 0) {
#ifdef DEBUG_FILE
				LOG_X("sector of file %s\n",
					unix_filename(pfl->file_id));
#endif
				strcpy((char *)&path, pfs->ppath);
				strcat((char *)&path, PATHSEPARATORSTRING);
				strcat((char *)&path, unix_filename(pfl->file_id));
				if ((fp = fopen(path, mode)) != NULL &&
					!fseek(fp, (long)(pfl->f_record * 252L),
					   SEEK_SET)) {
						bytes = fread(&sector_buffer[4],
							1, 252, fp);
					fclose(fp);
				// stuff last sector of file with 0
					for (i = 4+bytes; i < SECTOR_SIZE; i++)
						sector_buffer[i] = 0;
				} else { // unable to read sector
					drq = 0;
					str = 0x10;
					byteCount = 0;
				} // else
			} else {
			// new file with name tmpXX
#ifdef DEBUG_FILE
				LOG_X("sector of new file %s\n",
				  (pfs->pnew_file+(NEW_FILE1 - pfl->file_id))->filename);
				if ((tr == 0x15 && sr == 0x1d) ||
					(tr == 0x17 && sr == 0x19)) {
					LOG("     important\n");
				}
#endif
				fp = (pfs->pnew_file+(NEW_FILE1 - pfl->file_id))->fp;
				if (!fseek(fp, (long)(pfl->f_record * 252L), SEEK_SET)) {
					bytes = fread(&sector_buffer[4], 1, 252, fp);
					// stuff last sector of file with 0
					for (i = 4+bytes; i < SECTOR_SIZE; i++)
						sector_buffer[i] = 0;
					fseek(fp, 0L, SEEK_END); // position end of file
				} else { // unable to read sector
					drq = 0;
					str = 0x10;
					byteCount = 0;
				} // else
			} // else
			sector_buffer[0] = pfl->next.st.trk;
			sector_buffer[1] = pfl->next.st.sec;
			sector_buffer[2] = pfl->record_nr[0];
			sector_buffer[3] = pfl->record_nr[1];
			break;
	} // switch
} //read_nafs_sector


void E2floppy::write_nafs_sector(void)
{
	char			*p, path[PATH_MAX+1], *mode;
	Word	 		i, di;
	size_t			bytes;
	SWord			new_file_index;
	FILE			*fp;
	struct s_link_table 	*pfl;

#ifdef DEBUG_FILE
	LOG_XX("write: %02X/%02X ", tr, sr);
#endif
	fp    = NULL;
#ifdef sun
	mode  = "r+";
#else
	mode  = "rb+";
#endif
	bytes = 0;	// satisfy compiler
	i     = tr * MAX_SECTOR + (sr - 1);
	pfl   = pfs->pflex_links + i;
	switch (pfl->file_id) {
		case SYSTEM :
#ifdef DEBUG_FILE
			LOG("systemsector\n");
#endif
			p = NULL; // satisfy compiler
			if (sr >= 3)
				p = (char *)(pfs->pflex_sys_info + sr - 3);
			else if (sr == 2)
				p = (char *)pfs->pflex_unused;
			else if (sr == 1) {
#ifdef sun
				mode = "w";
#else
				mode = "wb";
#endif
				strcpy((char *)&path, pfs->ppath);
				strcat((char *)&path, PATHSEPARATORSTRING "boot");
				if ((fp = fopen(path, mode)) != NULL) {
					bytes = fwrite(&sector_buffer, 1,
						SECTOR_SIZE, fp);
					fclose(fp);
				}
				if (bytes != SECTOR_SIZE)
					str = 0x10;
				break;
			} else
				break;
			memcpy(p, sector_buffer, pfs->byte_p_sector);
			break;
		case DIRECTORY:
#ifdef DEBUG_FILE
			LOG("directorysector\n");
#endif
			di = pfl->f_record;
			p = (char *)(pfs->pflex_directory + di);
			check_for_delete(di, (s_dir_sector *)sector_buffer);
			check_for_new_file(di, (s_dir_sector *)sector_buffer);
			check_for_rename(di, (s_dir_sector *)sector_buffer);
			check_for_extend(di, (s_dir_sector *)sector_buffer);
			memcpy(p, sector_buffer, pfs->byte_p_sector);
			break;
		case FREE_CHAIN :
#ifdef DEBUG_FILE
			LOG("free chain\n");
#endif
			if (pfs->dir_extend.st.trk == tr &&
			   pfs->dir_extend.st.sec == sr) {
			   	extend_directory(pfs, i, (s_dir_sector *)sector_buffer);
#ifdef DEBUG_FILE
				LOG("      extend directory\n");
#endif
				break;
			} // if
			if (last_of_free_chain(tr, sr) &&
				(sector_buffer[1] || sector_buffer[0])) {
				pfl->next.st.trk  = sector_buffer[0];
				pfl->next.st.sec  = sector_buffer[1];
				pfl->record_nr[0] = sector_buffer[2];
				pfl->record_nr[1] = sector_buffer[3];
#ifdef DEBUG_FILE
				LOG("      file deleted\n");
#endif
				break;
			}
			if ((new_file_index = index_of_new_file(tr, sr)) < 0) {
#ifdef DEBUG_FILE
				LOG("   ** error: unable to create new file\n");
#endif
				str = 0x10; // no more new files can be created
				break;
			}
#ifdef DEBUG_FILE
			LOG_X("      file %s\n",
				(pfs->pnew_file+new_file_index)->filename);
#endif
			fp = (pfs->pnew_file+new_file_index)->fp;
			(pfs->pnew_file+new_file_index)->next.st.trk = sector_buffer[0];
			(pfs->pnew_file+new_file_index)->next.st.sec = sector_buffer[1];
			pfl->file_id  = NEW_FILE1 - new_file_index;
			// pfl->f_record = (pfs->pnew_file+new_file_index)->f_record++;
			//pfl->f_record = ((sector_buffer[2] << 8) | sector_buffer[3]) - 1;
			pfl->next.st.trk  = sector_buffer[0];
			pfl->next.st.sec  = sector_buffer[1];
			pfl->record_nr[0] = sector_buffer[2];
			pfl->record_nr[1] = sector_buffer[3];
			pfl->f_record = record_nr_of_new_file(new_file_index, i);
			if (ftell(fp) != (pfl->f_record * 252L) && 
			   fseek(fp, (long)(pfl->f_record * 252L), SEEK_SET) != 0)
				str = 0x10;
			else if ((bytes =
				fwrite(&sector_buffer[4], 1, 252, fp)) != 252)
				str = 0x10;
			// (pfs->pnew_file+new_file_index)->f_record++;
			break;
		default : 
			if (pfl->file_id >=  0) {
#ifdef DEBUG_FILE
				LOG_X("sector of file %s\n",
					unix_filename(pfl->file_id));
#endif
				strcpy((char *)&path, pfs->ppath);
				strcat((char *)&path, PATHSEPARATORSTRING);
				strcat((char *)&path, unix_filename(pfl->file_id));
				pfl->next.st.trk = sector_buffer[0];
				pfl->next.st.sec = sector_buffer[1];
				pfl->record_nr[0]= sector_buffer[2];
				pfl->record_nr[1]= sector_buffer[3];
				if ((fp = fopen(path, mode)) == NULL)
					str = 0x10;
				else {
					if (ftell(fp) != pfl->f_record && 
					   fseek(fp,
					     (long)(pfl->f_record * 252L),
					     SEEK_SET) != 0)
						str = 0x10;
					else if ((bytes =
						fwrite(&sector_buffer[4], 1, 252, fp)) != 252)
						str = 0x10;
					fclose(fp);
				} // else
			} else {
#ifdef DEBUG_FILE
				LOG_X("sector of new file %s\n",
					unix_filename(pfl->file_id));
#endif
				fp = (pfs->pnew_file+(NEW_FILE1 - pfl->file_id))->fp;
				if (ftell(fp) != pfl->f_record && 
				   fseek(fp, (long)(pfl->f_record * 252L),
				   	SEEK_SET) != 0)
					str = 0x10;
				else if ((bytes =
					fwrite(&sector_buffer[4], 1, 252, fp)) != 252)
					str = 0x10;
			} //else
			pfl->next.st.trk = sector_buffer[0];
			pfl->next.st.sec = sector_buffer[1];
			pfl->record_nr[0]= sector_buffer[2];
			pfl->record_nr[1]= sector_buffer[3];
	} // switch
} // write_nafs_sector


Byte E2floppy::init_nafs_drive(struct s_floppy *pfloppy)
{
	struct stat		sbuf;
	static Word		nafs_number = 0;

	if (pfloppy->ppath == NULL || (strlen(pfloppy->ppath) == 0))
		return 0;
	if(stat(pfloppy->ppath, &sbuf))
		return 0;
#ifdef _MSC_VER
	if (sbuf.st_mode & _S_IFDIR) {
#else
	if (S_ISDIR(sbuf.st_mode)) {
#endif
	  	mount_nafs(pfloppy, nafs_number);
	  	nafs_number++;
	  	return 1;
	} // if
	return 0;
} // init_nafs_drive

void E2floppy::mount_nafs(struct s_floppy *pfloppy, Word index)
{
	int wp; // should be int because of call 'access'

	wp = access(pfloppy->ppath, W_OK);
  	header_for_nafs_format(pfloppy, wp);
	initialize_new_file_table(pfloppy);
  	initialize_flex_sys_info_sectors(pfloppy, index);
  	initialize_flex_unused_sector(pfloppy);
  	fill_flex_directory(pfloppy, wp);
  } // mount_nafs


void E2floppy::umount_nafs(Word index)
{
	close_new_files(index);
	free_nafs_memory(index);
}  // umount_nafs

#endif // #ifdef NAFS
